@class CheckBox (Core)	@file=CheckBox

@ The |CheckBox| widget contains a text label and a rectangular box
that is either unchecked (empty) or checked (contains an x). The current
state of the checkbox can be obtained by querying the |state| resource.


@PUBLIC

@ The text to be displayed

	@var String label = NULL

@ The text is drawn using the font given in the |font| resources.

	@var <FontStruct> XFontStruct *font = <String> XtDefaultFont

@ The |color| resource specifies the color used for drawing the text.

	@var Pixel color = <String> XtDefaultForeground

@ The offset from the text to the outside edge of the widgets window.

        @var Dimension internalOffset = 2

@ The size of the checkbox.

        @var Dimension boxSize = 11

@ The number of pixels between the checkbox and the text.

        @var Dimension boxGap = 9

@ The state of the checkbox. True means on, False means off.

        @var Boolean state = False

@ The |border_width| by default is 0.

        @var border_width = 0

@

@PRIVATE

@ The graphics context |gc| gets modified each time a color changes or a font changes
	int the text.

	@var GC gc

@ For graying out the text, another GC is used.

	@var GC graygc


@translations

@ The |toggle| action toggles the widget between `on' and `off'. By
default it is bound to a click of the left mouse button as well as to
the Space key.

        @trans <Btn1Down>,<Btn1Up>: toggle() 
        @trans <Key>space: toggle()

@actions                                                                        

@ The |toggle| action switches the state of the checkbox.

@proc toggle
{
    XtVaSetValues($, XtNstate, !$state, NULL);
}

@CLASSVARS
	@var compress_exposure =
		XtExposeCompressMultiple|XtExposeGraphicsExposeMerged


@UTILITIES


@ create the GC used for drawing text.

@proc make_gc($)
{
	XtGCMask mask, dynmask, dontcaremask;
	XGCValues values;

	values.foreground=$color;
	values.background=$background_pixel;
	values.font=$font->fid;
	dynmask = GCForeground | GCBackground | GCFont;
	dontcaremask = GCJoinStyle | GCFillRule | GCTile
		| GCTileStipXOrigin | GCTileStipYOrigin | GCDashOffset | GCDashList
		| GCStipple | GCGraphicsExposures | GCArcMode | GCLineWidth | GCClipMask;
	mask=0;
	$gc=XtAllocateGC($,$depth,mask,&values,dynmask,dontcaremask);
}	

@ The |make_graygc| routine creates a GC for graying out the text. It
contains a stipple in the background color, that will be applied over
the text.

@proc make_graygc($) {
    XtGCMask mask;
    XGCValues values;

    if ($graygc != NULL) XtReleaseGC($, $graygc);
    values.foreground = $background_pixel;
    values.stipple =
        XCreateBitmapFromData(XtDisplay($),
                              RootWindowOfScreen(XtScreen($)),
                              stip4_bits, stip4_width, stip4_height);
    values.fill_style = FillStippled;
    mask = GCForeground | GCStipple | GCFillStyle;
    $graygc = XtGetGC($, mask, &values);
	}


@ This function is used to calculate the maximum ascender and descender
	for a line of text chunks.  This could be the ascent from one font and
	the descent from a different font to give the maximum line height.

@proc compute_line_height($, int *ma, int *md)
{
	XFontStruct *f;

	*ma=0, *md=0;
	f=$font;
	*ma =  $boxSize > f->ascent ? $boxSize : f->ascent;
	*md =  f->descent;
}

@ Calculate the size of the widget based on the text that it has to display.

@proc compute_size($, Dimension *wd, Dimension *ht)
{
	int ma, md;
	
	compute_line_height($, &ma, &md);
	*ht = ($internalOffset)*2 + ma + md;
		
	*wd = XTextWidth($font,$label,strlen($label)) + ($internalOffset)*2 
			+ $boxSize + $boxGap;
}


@METHODS

@ The |initialize| method sets the private variables to default values.

@proc initialize
{
    if ($label) {
		$label = XtNewString($label);
		compute_size($, &$width, &$height);
		}
	$gc=NULL;
	$graygc=NULL;
	make_gc($);
	make_graygc($);
}


@proc destroy
{
	XtReleaseGC($, $gc);
	XtFree($label);
}


@ The |set_values| method copies all changed string resources.

@proc set_values
{
	Boolean do_size = False;
	Boolean need_redisplay = False;

	if ($background_pixel != $old$background_pixel)
		make_graygc($);

	if($old$label != $label) {
        XtFree($old$label);
        $label = XtNewString($label);
		do_size = True;
		}

    if ($state != $old$state) 
		do_size = True;
    if ($boxSize != $old$boxSize) 
		do_size = True;
    if ($boxGap != $old$boxGap) 
		do_size = True;

	if($old$internalOffset != $internalOffset)
		do_size = True;

	if($old$font != $font)
		do_size = True;

	if (($old$color != $color) ||
		($old$sensitive != $sensitive))
			need_redisplay = True;

		need_redisplay = True;
		
	if(do_size) {
		compute_size($, &($width), &($height));
		need_redisplay = True;
		}

	return need_redisplay;
}

@ The |query_geometry| method is called when the parent wants to resize the 
widget.
	
@proc query_geometry {
	return XtGeometryYes;
	}

@ The |expose| method is called if the widget was overlapped and became
visible again.  There are some basic checks to see if certain
chunks need to be displayed. ie: are they above the exposed region or
below it then they don't need to be displayed.  And if the remainder
are partially or entirely within the region then they need to be
displayed.
	
@proc expose {
	XFontStruct *f;
    int ma, md;
	int ypos;
	int x1, y1, x2, y2;
		
	if(! XtIsRealized($))
		return ;

	if(event == NULL)
	{
		XClearWindow(XtDisplay($),XtWindow($));
		return;
	}

	/* printf("Expose: width=%d\n", $width);*/

    compute_line_height($, &ma, &md);
	ypos=ma + $internalOffset;

	/* draw the checkbox */
	XDrawRectangle(XtDisplay($), XtWindow($), $gc, 
		$internalOffset, $internalOffset, $boxSize, $boxSize);
	if ($state) {
		x1=$internalOffset;
		y1=$internalOffset;
		x2=x1+$boxSize;
		y2=y1+$boxSize;
		XDrawLine(XtDisplay($), XtWindow($), $gc, x1, y1, x2, y2);
		XDrawLine(XtDisplay($), XtWindow($), $gc, x1, y2, x2, y1);
		}

	/* draw the label */
	f = $font;
	XSetFont(XtDisplay($), $gc, f->fid);
	XSetForeground(XtDisplay($), $gc, $color);
			
	XDrawString(XtDisplay($),XtWindow($),$gc,
			$internalOffset+$boxSize+$boxGap, ypos, $label, strlen($label));

	/* Gray out if not sensitive */
	if (! $sensitive) {
		XFillRectangle(XtDisplay($), XtWindow($), $graygc, 0,
			0, $width, $height);
		}
	}


@IMPORTS

@incl "stip4.bm"
@incl <assert.h>
@incl <ctype.h>
@incl <stdio.h>
@incl <string.h>
