# Mlabel
# Author: Lachlan Wetherall
# Modified from the Label widget from the Free Widget Foundation
# which was written by Bert Bos <bert@let.rug.nl>

@class Mlabel (Composite) @file=Mlabel

@ The Label class has the capability to display one or more lines of
text in a single font. Otherwise it is the same as the Board class.
The text can be left, right or center justified and it can be centered
vertically or put against the top or the bottom of the widget. There
is also a resource to set tab stops.

There are two ways of highlighting portions of the label: reversing
the colors or changing the foreground. Both methods can be combined,
though the result when both highlighting methods are applied to the
same part of the text is undefined.


@public

@ The text is a single string, which may contain embedded newlines.
There is no provision for changing fonts in the middle of a text.

	@var String label = NULL

@ The text is drawn in the font which is given as the |font| resource.

	@var <FontStruct> XFontStruct *font = <String> XtDefaultFont

@ The foreground color is the color used to draw the
text. |hlForeground| is the foreground for highlighted text.

	@var Pixel foreground = <String> XtDefaultForeground
	@var Pixel hlForeground = <String> XtDefaultForeground

@ The text can be aligned in the widget in nine ways: left, right or
center, combined with top, center or bottom. Symbolic constants
|XfwfTop|, |XfwfBottom|, |XfwfLeft| and |XfwfRight| can be added together to
get the desired alignment.  The alignment is actually a four-bit
number made up of two parts of 2 bits added together: 1 is left, 2 is
right, 0 is center, 4 is top, 8 is bottom, 0 is vertical center. Thus
5 (= 1 + 4) means top left and 2 (= 2 + 0) means center right. For
easier specification, there is also a converter from strings, that
accepts string like `top left' or `center right'.

	@var Alignment alignment = 0

@ The |topmargin| is only used when the text is not centered. It gives
the number of pixels between the frame and the top of the text.

	@var Dimension topMargin = 2

@ The |bottomMargin| is only used to compute the preferred size of the
button in case |shrinkToFit = True|.

	@var Dimension bottomMargin = 2

@ The |leftMargin| is only used when the text is not centered. It
gives the number of pixels between the frame and the left edge of the
text, and if possible also between the frame and the right edge of the
text.

	@var Dimension leftMargin = 2

@ The |rightMargin| is only used to compute the preferred size of the
button in case |shrinkToFit = True|.

	@var Dimension rightMargin = 2

@ Buttons will normally not occupy the full area of their parents.
Most likely they will be a fixed size or a size depending on the
label. By setting the |shrinkToFit| resource to True, the width and
height are recomputed with every new label.

	@var Boolean shrinkToFit = False

@ It is possible to set a part of the label apart by drawing it in
reverse. The |rvStart| resource gives the index of the first
character to draw in reverse video.

	@var int rvStart = 0

@ The |rvLength| resource contains the number of characters to
draw in reverse video.

	@var int rvLength = 0

@ The start and length of the highlighted portion are set with the
resources |hlStart| and |hlLength|.

	@var int hlStart = 0
	@var int hlLength = 0


@private

@ For faster drawing, the number of lines in the text is stored in a
private variable by the |set_values| and |initialize| methods.

	@var int nlines

@ For drawing the text, this GC is used.

	@var GC gc

@ This GC is for the text that is drawn in reverse video.

	@var GC rv_gc

@ The GC for the highlighted portion of the text

	@var GC hl_gc

@ When the |shrinkToFit| resource is set, we need the minimum area
necessary for the complete label to be visible. |label_width| and
|label_height| include the size of |margin|.

	@var Dimension label_width
	@var Dimension label_height

@exports

@ The type |Alignment| is actually an integer, but it is given a
different name to allow a type converter to be installed for it.

        @type Alignment = int

@ The symbolic constants can be added together to form an alignment.
Various widgets use this to position labels, other widgets, etcetera.

        @def XfwfCenter = 0
        @def XfwfLeft = 1
        @def XfwfRight = 2
        @def XfwfTop = 4
        @def XfwfBottom = 8


@methods


@ When the text or the font change, the
private variables |nlines|, |label_height| and |label_width| are
updated.

|need_count| is set to |True| if the size of the label changes.
|need_count| implies |need_redisplay|.

@proc set_values
{
    Boolean need_redisplay = False, need_count = False;

    if ($font != $old$font) {
		make_gc($);
		if ($label != NULL) need_count = True;
	    }

    if ($foreground != $old$foreground
	 	|| $hlForeground != $hlForeground
		|| $background_pixel != $old$background_pixel) {
		make_gc($);
		if ($label != NULL) need_redisplay = True;
	    }

    if ($topMargin != $old$topMargin
		|| $bottomMargin != $old$bottomMargin
		|| $leftMargin != $old$leftMargin
		|| $rightMargin != $old$rightMargin)
		need_count = True;

    if ($sensitive != $old$sensitive)
		if ($label != NULL) need_redisplay = True;

    if ($rvStart != $old$rvStart || $rvLength != $old$rvLength
	 	|| $hlStart != $old$hlStart || $hlLength != $old$hlLength)
			if ($label != NULL) need_redisplay = True;

    if ($label != $old$label) {
		XtFree($old$label);
		$label = XtNewString($label);
		need_count = True;
    	}

    if (need_count) {
		count_lines($);
		need_redisplay = True;
	    }

    if (need_count && $shrinkToFit) {
		$width = $label_width;
		$height = $label_height;
	    need_redisplay = True;
		}

    return need_redisplay;
}

@ The |initialize| methods creates the first GC's and initializes the
private variables. It sets the GC's to |NULL| and calls two utility
routines to actually create them.

@proc initialize
{
    if ($label) $label = XtNewString($label);
    count_lines($);
    $gc = NULL;
    $rv_gc = NULL;
    $hl_gc = NULL;
    make_gc($);
    if ($shrinkToFit) {
		$width = $label_width;
		$height = $label_height;
	    }
}

@ The |expose| method is responsible for drawing the text. The text is
put in the position given in |alignment|. The text is always kept
within the frame. If necessary, the text is clipped. The routine ends
by calling the |expose| method from the superclass, which is
responsible for drawing the frame.

The part of the text that is to appear in reverse video is drawn with
the |rv_gc| GC.

@def max(a, b) = ((a) > (b) ? (a) : (b))

@def min(a, b) = ((a) < (b) ? (a) : (b))

@def draw_line(dpy, win, from, to) =
    do {
	if ($hlStart >= to) hstart = to;
 	else hstart = max($hlStart, from);
	if ($hlStart + $hlLength <= from) hend = hstart;
 	else hend = min($hlStart + $hlLength, to);
        if ($rvStart >= to) rstart = to;
	else rstart = max($rvStart, from);
	if ($rvStart + $rvLength <= from) rend = rstart;
	else rend = min($rvStart + $rvLength, to);
	w1 = XTextWidth($font, $label + from, rstart - from);
	w2 = XTextWidth($font, $label + rstart, rend - rstart);
	w3 = XTextWidth($font, $label + rend, to - rend);
 	w4 = XTextWidth($font, $label + hstart, hend - hstart);
 	w5 = XTextWidth($font, $label + from, hstart - from);
	if ($alignment & XfwfLeft)
	    x = rect.x;
	else if ($alignment & XfwfRight)
	    x = rect.x + rect.width - w1 - w2 - w3;
	else
	    x = rect.x + (rect.width - w1 - w2 - w3)/2;
	if (w1)
	    XDrawString(dpy, win, $gc, x, y, $label + from,
			     rstart - from);
	if (w2)
	    XDrawImageString(dpy, win, $rv_gc, x + w1, y, $label
			     + rstart, rend - rstart);
	if (w3)
	    XDrawString(dpy, win, $gc, x + w1 + w2, y, $label +
			     rend, to - rend);
 	if (w4)
 	    XDrawString(dpy, win, $hl_gc, x + w5, y, $label
 			     + hstart, hend - hstart);
    } while (0)

@proc expose
{
    Region reg;
    XRectangle rect;
    int baseline;
    int w1, w2, w3, w4, w5;
    int x, y, i, j, rstart, rend, hstart, hend;

    if (! XtIsRealized($)) return;
    if ($label != NULL) {
	baseline = $font->ascent + $font->descent;
	rect.x=$leftMargin;  
	rect.width = $width -($leftMargin + $rightMargin);
	rect.y = $topMargin;
	rect.height = $height - ($topMargin + $bottomMargin);
	reg = XCreateRegion();
	XUnionRectWithRegion(&rect, reg, reg);
	if (region != NULL) XIntersectRegion(region, reg, reg);
	XSetRegion(XtDisplay($), $gc, reg);
	XSetRegion(XtDisplay($), $rv_gc, reg);
 	XSetRegion(XtDisplay($), $hl_gc, reg);
	if ($alignment & XfwfTop)
	    y = rect.y + $font->ascent;
	else if ($alignment & XfwfBottom)
	    y = rect.y + rect.height - $nlines * baseline + $font->ascent;
	else
	    y = rect.y + (rect.height - $nlines * baseline)/2 + $font->ascent;
	for (i = 0, j = 0; $label[i]; i++) {
	    if ($label[i] == '\n') {
		draw_line(XtDisplay($), XtWindow($), j, i);
		j = i + 1;
		y += baseline;
	    }
	}
	draw_line(XtDisplay($), XtWindow($), j, i);

	XSetClipMask(XtDisplay($), $gc, None);
	XSetClipMask(XtDisplay($), $rv_gc, None);
 	XSetClipMask(XtDisplay($), $hl_gc, None);
    }
}



@utilities

@ The |make_gc| routine creates the GCs for the normal and highlighted
text.

@proc make_gc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($gc != NULL) XtReleaseGC($, $gc);
    values.background = $background_pixel;
    values.foreground = $foreground;
    values.font = $font->fid;
    mask = GCFont | GCBackground | GCForeground;
    $gc = XtGetGC($, mask, &values);

    if ($rv_gc != NULL) XtReleaseGC($, $rv_gc);
    values.foreground = $background_pixel;
    values.background = $foreground;
    values.font = $font->fid;
    mask = GCFont | GCBackground | GCForeground;
    $rv_gc = XtGetGC($, mask, &values);

    if ($hl_gc != NULL) XtReleaseGC($, $hl_gc);
    values.background = $background_pixel;
    values.foreground = $hlForeground;
    values.font = $font->fid;
    values.function = GXcopy;
    $hl_gc = XtGetGC($, mask, &values);
}

@ The funtion |count_lines| computes the correct values for the
private variables |nlines|, |label_width| and |label_height|.

@proc count_lines($)
{
    String p, s;
    int w;

    $nlines = 0;
    $label_width = 0;
    if ($label) {
	for (p = $label, $nlines = 1, s = $label; *s; s++) {
	    if (*s == '\n') {
		$nlines++;
		w = XTextWidth($font, p, s - p);
		p = s + 1;
		if (w > $label_width) $label_width = w;
	    }
	}
	w = XTextWidth($font, p, s - p);
	if (w > $label_width) $label_width = w;
    }
    $label_height = $nlines * ($font->ascent + $font->descent);
    $label_width += $leftMargin + $rightMargin;
    $label_height += $topMargin + $bottomMargin;
}

@imports

@incl <stdio.h>
