@class RadioBtn (Core)	@file=RadioBtn

@ The |RadioBtn| widget contains a number of text labels each adjacent to a
radio push button.  Only one button at a time can be selected.


@PUBLIC

@ The text for all the radio buttons are specifed by the |buttons| 
string resource.  The text for separate buttons is delimited by a 
vertical bar.

	@var String buttons = NULL

@ The text is drawn using the font given in the |font| resource.

	@var <FontStruct> XFontStruct *font = <String> XtDefaultFont

@ The |color| resource specifies the color used for drawing the text.

	@var Pixel color = <String> XtDefaultForeground

@ The offset from the text to the outside edge of the widgets window.

        @var Dimension internalOffset = 2

@ The size of each radio button.

        @var Dimension circleSize = 11

@ The number of pixels between the radio button and the text.

        @var Dimension circleGap = 9

@ The number of pixels between each button.

        @var Dimension verticalGap = 9

@ Which button is selected.

        @var int selected = 0

@ The |radioCallback| is called from the |press| action if a different
button was selected. The |call_data| argument will be an integer indicating
the new selection.

	@var <Callback> XtCallbackList radioCallback = NULL

@ The border width is 0 by default.

        @var border_width = 0

@PRIVATE

@ The graphics context |gc| gets modified each time a color changes or a 
font changes in the text.

	@var GC gc


@translations

@ The |press| action selects the button underneath the mouse cursor. By
default it is bound to a click of the left mouse button.

        @trans <Btn1Down>,<Btn1Up>: press() 

@actions                                                                        

@ The |press| action selects the radio button under the mouse cursor. 

@proc press
{
    int my, ma, md, v, s;

    my=event->xbutton.y;
    compute_line_height($, &ma, &md);
	v=ma + md + $verticalGap;

	s= (my-$internalOffset)/v;
	if (s!=$selected)
		XtCallCallbackList($, $radioCallback, (XtPointer) s);
	XtVaSetValues($, XtNselected, s, NULL);
	}

@CLASSVARS
	@var compress_exposure =
		XtExposeCompressMultiple|XtExposeGraphicsExposeMerged


@UTILITIES


@ Create the GC used for drawing text.

@proc make_gc($)
{
	XtGCMask mask, dynmask, dontcaremask;
	XGCValues values;

	values.foreground=$color;
	values.background=$background_pixel;
	values.font=$font->fid;
	dynmask = GCForeground | GCBackground | GCFont;
	dontcaremask = GCJoinStyle | GCFillRule | GCTile
		| GCTileStipXOrigin | GCTileStipYOrigin | GCDashOffset | GCDashList
		| GCStipple | GCGraphicsExposures | GCArcMode | GCLineWidth | GCClipMask;
	mask=0;
	$gc=XtAllocateGC($,$depth,mask,&values,dynmask,dontcaremask);
}	

@ This function is used to calculate the maximum ascender and descender
for a line of text chunks.  This could be the ascent from one font and
the descent from a different font to give the maximum line height.

@proc compute_line_height($, int *ma, int *md)
{
	XFontStruct *f;

	*ma=0, *md=0;
	f=$font;
	*ma =  $circleSize > f->ascent ? $circleSize : f->ascent;
	*md =  f->descent;
}

@ Calculate the size of the widget based on the text that it has to display.

@proc compute_size($, Dimension *wd, Dimension *ht)
{
	int ma, md, num_buttons, w, maxw;
	char *cp, *start, *end;
	
	if ($buttons != NULL) {
		num_buttons = 1;
		for (cp=$buttons; *cp!='\0'; cp++)
			if (*cp=='|') num_buttons++;
		}
	compute_line_height($, &ma, &md);
	*ht = ($internalOffset)*2 + (ma + md)*num_buttons + 
			$verticalGap*(num_buttons-1);
		
	start=$buttons;
	end=$buttons; 
	while ((*end != '|') && (*end != '\0'))	end++;
	
	/* find maximum button label length */
	maxw=0;
	while (*start != '\0') {
		w = XTextWidth($font,start,end-start); 
		if (w>maxw) maxw=w;

		start=end;
		if (*start == '|') start++;
		end=start;
		while ((*end != '|') && (*end != '\0'))	end++;
		}
	*wd = maxw + ($internalOffset)*2 + $circleSize + $circleGap;
}


@METHODS

@ The |initialize| method sets the private variables to default values.

@proc initialize
{
    if ($buttons) {
		$buttons = XtNewString($buttons);
		compute_size($, &$width, &$height);
		}
	$gc=NULL;
	make_gc($);
}


@proc destroy
{
	XtReleaseGC($, $gc);
	XtFree($buttons);
}


@ The |set_values| method copies all changed string resources.

@proc set_values
{
	Boolean do_size = False;
	Boolean need_redisplay = False;

	if($old$buttons != $buttons) {
        XtFree($old$buttons);
        $buttons = XtNewString($buttons);
		do_size = True;
		}

    if ($selected != $old$selected) 
		do_size = True;
    if ($circleSize != $old$circleSize) 
		do_size = True;
    if ($circleGap != $old$circleGap) 
		do_size = True;

	if($old$internalOffset != $internalOffset)
		do_size = True;

	if($old$font != $font)
		do_size = True;
		
	if($old$color != $color)
		need_redisplay = True;

	if(do_size) {
		compute_size($, &($width), &($height));
		need_redisplay = True;
		}

	return need_redisplay;
}

@ The |query_geometry| method is called when the parent wants to resize the 
widget.
	
@proc query_geometry {
	return XtGeometryYes;
	}

@ The |expose| method is called if the widget was overlapped and became
	visible again.  There are some basic checks to see if certain
	chunks need to be displayed. ie: are they above the exposed region or
	below it then they don't need to be displayed.  And if the remainder
	are partially or entirely within the region then they need to be
	displayed.
	
@proc expose
{
	XFontStruct *f;
    int ma, md;
	int x, y;
	int btn_num;
	char *start, *end;
	
	if(! XtIsRealized($))
		return ;

	if(event == NULL)
	{
		XClearWindow(XtDisplay($),XtWindow($));
		return;
	}

	/* printf("Expose: width=%d\n", $width);*/

    compute_line_height($, &ma, &md);

	f = $font;
	XSetFont(XtDisplay($), $gc, f->fid);
	XSetForeground(XtDisplay($), $gc, $color);

	start=$buttons;
	end=$buttons; 
	while ((*end != '|') && (*end != '\0'))	end++;
	
	btn_num=0;
	while (*start != '\0') {
		/* draw the circle */
		x = $internalOffset;
		y = $internalOffset + btn_num * (ma + md + $verticalGap);
		XDrawArc(XtDisplay($), XtWindow($), $gc, x, y, $circleSize, $circleSize, 0, 360*64);
		if (btn_num==$selected)
			XFillArc(XtDisplay($), XtWindow($), $gc, x+2, y+2, $circleSize-4, $circleSize-4, 0, 360*64);

		/* draw the label */
			
		x = $internalOffset+$circleSize+$circleGap;
		y = $internalOffset + btn_num * (ma + md + $verticalGap) + ma;
		XDrawString(XtDisplay($),XtWindow($),$gc,
			x, y, start, end-start);

		btn_num++;
		start=end;
		if (*start == '|') start++;
		end=start;
		while ((*end != '|') && (*end != '\0'))	end++;
		}
}


@IMPORTS

@incl <assert.h>
@incl <ctype.h>
@incl <stdio.h>
@incl <string.h>
